from typing import List

from transformers import PreTrainedTokenizerFast

from briton.schema import ModelInput, get_prompt


async def collect_text(async_text_iter) -> tuple[str, int]:
    full_text = ""
    completion_tokens = 0
    async for delta in async_text_iter:
        completion_tokens += len(delta.output_ids)
        full_text += delta.output_text
    return full_text, completion_tokens



import asyncio
import logging
from typing import Any, Dict, List, Optional, Tuple, cast

import fastapi
from transformers import PreTrainedTokenizerFast

from briton.backend.backend_types import (
    BackendConfig,
    BatchSchedulerPolicy,
    InferBackend,
    LazyLoadParams,
    LoadParams,
)
from briton.backend.default_backend import DefaultBackend
from briton.backend.legacy_backend import LegacyBackend
from briton.briton import BritonInteractor, BritonInteractorImpl
from briton.config_utils import trtllm_config_check
from briton.constants import (
    DEFAULT_BRITON_PORT,
    DEFAULT_MAX_FSM_WORKERS,
    OPENAI_COMPATIBLE_TAG,
    TOOL_CALL_IDS,
    TOOL_CALL_TOKENS,
)
from briton.fsm_cache import FsmCache, add_schema_to_cache
from briton.request_id_generator import RequestIdGenerator
from briton.schema import validate_model_input
from briton.secrets import get_hf_token_or_none
from briton.trtllm_config import TRTLLMConfiguration, TrussTRTLLMBatchSchedulerPolicy
from briton.truss_monitor import start_monitor

logger = logging.getLogger(__name__)


class Model:
    def __init__(self, **kwargs):
        self._loaded = False
        self._data_dir = kwargs["data_dir"]
        self._secrets = kwargs["secrets"]

        self._request_id_generator = RequestIdGenerator()
        config = kwargs["config"]
        model_metadata = config.get("model_metadata", {})
        trtllm_config_check(config)
        trtllm_config = TRTLLMConfiguration(**config.get("trt_llm"))
        self._backend_config = _generate_backend_config(model_metadata, trtllm_config)

        self._hf_token = get_hf_token_or_none(self._secrets)
        self._lazy_init_done = False
        self._lazy_init_lock = None

        # Allow passing briton_interactor for ease of testing
        self._briton_interactor: BritonInteractor = model_metadata.get(
            "briton_interactor", BritonInteractorImpl()
        )

        # Supports passing in a backend via model_metadata
        self._backends: List[InferBackend] = model_metadata.get("infer_backends", [])
        if self._backend_config.is_openai_compatible:
            self._backends.append(DefaultBackend())
        else:
            self._backends.append(LegacyBackend())

    def load(self):
        if self._loaded:
            return

        tokenizer, added_tokens = _build_tokenizer(
            tokenizer_repo=self._backend_config.tokenizer_repo,
            hf_token=self._hf_token,
            auto_tokenizer_from_pretrained=self._briton_interactor.auto_tokenizer_from_pretrained,
        )

        # We can't do this in __init__ because tokenizer is created above in
        # load function.
        _check_or_clear_backend_config_tool_call_token(self._backend_config, tokenizer)

        self._fsm_cache = FsmCache(
            self._briton_interactor.fsm_cache_dir(),
            tokenizer,
            self._backend_config.max_fsm_workers,
            self._backend_config.tool_call_token_id,
        )

        bcfg: BackendConfig = self._backend_config
        self._briton_interactor.load(
            model_name="briton",
            engine_path=self._data_dir,
            hf_tokenizer=bcfg.tokenizer_repo,
            work_dir=self._data_dir,
            fsm_cache_dir=self._fsm_cache.cache_dir,
            kv_cache_free_gpu_mem_fraction=bcfg.kv_cache_gpu_mem_fraction,
            port=DEFAULT_BRITON_PORT,
            added_tokens=added_tokens,
            max_num_tokens=bcfg.max_num_tokens,
            enable_chunked_context=bcfg.enable_chunked_context,
            hf_token=self._hf_token,
            tp_count=bcfg.tp,
            batch_scheduler_policy=bcfg.batch_scheduler_policy.value,
        )

        for backend in self._backends:
            backend.load(
                LoadParams(
                    generate_request_id=self._request_id_generator,
                    tokenizer=tokenizer,
                    config=self._backend_config,
                )
            )

        self._loaded = True

    async def predict(self, model_input: Dict[str, Any], request: fastapi.Request):
        """
        Run inference

        Note that the async nature of this function is a little tricky. Care is
        needed to make sure this function is a regular async function and not an
        async generator, i.e. there shouldn't be any direct yields in this
        function. This is because we need to support both streaming and
        non-streaming cases in this function. We do this by either returning an
        async-generator for the streaming case, or directly the full text for
        the other case. Returning an async generator for non-streaming case
        interferes with the open ai client proxy.
        """

        async def is_cancelled_fn():
            disconnected = await request.is_disconnected()
            if disconnected:
                logger.info("Request disconnected, cancelling.")
            return disconnected

        if not self._lazy_init_done:
            # While this isn't completely safe, the async lock needs to be
            # created within the same async loop where it will be used. Ideally,
            # the proper solution would involve supporting asynchronous load
            # function, but that is not currently supported in Truss. The risk is
            # that multiple initial requests could end up with different lock
            # instances, making the lock ineffective. In practice, this is
            # highly unlikely. This issue could occur if one request executes
            # the line below and then gets preempted, allowing another request
            # to execute the same line. However, since there is no async
            # operation in the following line, it is very unlikely for the
            # request to be preempted at that point.
            if self._lazy_init_lock is None:
                self._lazy_init_lock = asyncio.Lock()

            async with self._lazy_init_lock:
                stub = self._briton_interactor.create_grpc_stub(DEFAULT_BRITON_PORT)
                await start_monitor(self._briton_interactor, self.predict, logger)

                for backend in self._backends:
                    await backend.lazy_load(LazyLoadParams(briton_stub=stub))
                self._lazy_init_done = True

        validated_input = validate_model_input(
            model_input=model_input, supports_tools=self._backend_config.tool_call_token is not None
        )

        async def _add_schema_to_cache(schema: dict) -> str:
            return await add_schema_to_cache(self._fsm_cache, schema)

        # Firt backend that accepts the request gets it
        for backend in self._backends:
            req_details = await backend.accepts_request(validated_input)
            if req_details is not None:
                # todo implement total input tokens limit here using stats
                # one issue here is including current request tokens
                # todo handle asyncio.CancelledError
                return await backend.infer(
                    model_input=validated_input,
                    is_cancelled=is_cancelled_fn,
                    add_schema_to_cache=_add_schema_to_cache,
                    request_details=req_details,
                )


def _convert_batch_scheduler_policy(
    policy: TrussTRTLLMBatchSchedulerPolicy,
) -> BatchSchedulerPolicy:
    if policy == TrussTRTLLMBatchSchedulerPolicy.MAX_UTILIZATION:
        return BatchSchedulerPolicy.MAX_UTILIZATION

    if policy == TrussTRTLLMBatchSchedulerPolicy.GUARANTEED_NO_EVICT:
        return BatchSchedulerPolicy.GUARANTEED_NO_EVICT

    logger.warning(f"Unknown batch scheduler policy: {policy}. Using GUARANTEED_NO_EVICT.")
    return BatchSchedulerPolicy.GUARANTEED_NO_EVICT


def _supports_tool_calls(tool_call_token: str, tokenizer: PreTrainedTokenizerFast) -> bool:
    if not tool_call_token:
        return False

    tool_call_token_id = tokenizer.convert_tokens_to_ids(tool_call_token)
    return tool_call_token_id != tokenizer.unk_token_id


def _check_or_clear_backend_config_tool_call_token(
    backend_config: BackendConfig, tokenizer: PreTrainedTokenizerFast
):
    if not _supports_tool_calls(backend_config.tool_call_token, tokenizer):
        backend_config.tool_call_token = None
        backend_config.tool_call_token_id = None


def _generate_backend_config(
    model_metadata: dict,
    tllm_config: TRTLLMConfiguration,
):

    trtllm_build_config = tllm_config.build
    trtllm_runtime_config = tllm_config.runtime

    tags = model_metadata.get("tags", [])
    enable_kv_cache_reuse = trtllm_build_config.plugin_configuration.use_paged_context_fmha
    enable_chunked_context = trtllm_runtime_config.enable_chunked_context
    batch_scheduler_policy = _convert_batch_scheduler_policy(
        trtllm_runtime_config.batch_scheduler_policy
    )

    base_model = trtllm_build_config.base_model
    tool_call_token = TOOL_CALL_TOKENS.get(base_model)
    tool_call_token_id = TOOL_CALL_IDS.get(base_model)
    is_openai_compat = OPENAI_COMPATIBLE_TAG in tags

    return BackendConfig(
        model_metadata=model_metadata,
        is_openai_compatible=is_openai_compat,
        base_model=base_model,
        tp=trtllm_build_config.tensor_parallel_count,
        tokenizer_repo=trtllm_build_config.checkpoint_repository.repo,
        kv_cache_gpu_mem_fraction=trtllm_runtime_config.kv_cache_free_gpu_mem_fraction,
        enable_kv_cache_reuse=enable_kv_cache_reuse,
        enable_chunked_context=enable_chunked_context,
        max_num_tokens=trtllm_build_config.max_num_tokens,
        max_seq_len=trtllm_build_config.max_seq_len,
        batch_scheduler_policy=batch_scheduler_policy,
        default_max_tokens=trtllm_runtime_config.request_default_max_tokens,
        max_fsm_workers=DEFAULT_MAX_FSM_WORKERS,
        tool_call_token=tool_call_token,
        tool_call_token_id=tool_call_token_id,
    )


def _build_tokenizer(
    tokenizer_repo: str,
    auto_tokenizer_from_pretrained: callable,
    hf_token: Optional[str] = None,
) -> Tuple[PreTrainedTokenizerFast, List[str]]:
    # TODO(pankaj) Support loading bundled tokenizer rather than from HF
    raw_tokenizer = auto_tokenizer_from_pretrained(tokenizer_repo, hf_token=hf_token)
    tokenizer = cast(PreTrainedTokenizerFast, raw_tokenizer)
    # We only support Llama and mistral with Briton, for which this should
    # apply.
    assert isinstance(tokenizer, PreTrainedTokenizerFast)

    # These are tokens outside of tokenizer.json. We need to pass these to
    # Briton, to pass to rust tokenizer.
    added_token_decoders = tokenizer.added_tokens_decoder
    added_tokens = list(added_token_decoders.values())

    return tokenizer, added_tokens



import asyncio
from logging import Logger
from typing import Awaitable, Callable, Optional

import aiohttp
from starlette.requests import Request as StarletteRequest

from briton.async_util import predicate_with_timeout, retry_predicate
from briton.briton import BritonInteractor
from briton.constants import LOCAL_PREDICT_ENDPOINT
from briton.proc_utils import kill_current_process_and_children


async def monitor(
    monitor_fn: Callable[[], Awaitable[bool]],
    on_fail: Callable[[], None],
    should_continue: Optional[Callable[[], bool]],
    period_secs: int = 60,
    max_retries: int = 3,
    timeout_secs: int = 60,
    retry_delay_secs: int = 2,
):
    """Monitor by calling a function.
    Args:
        monitor_fn: The function to call to check.
        on_fail: Function to call when the model is unresponsive.
        should_continue: Function to determine if monitoring should continue.
        period_secs: How often to check the model.
        retries: How many times to retry.
    """
    if should_continue is None:
        should_continue = lambda: True

    while should_continue():
        if not await _check_with_timeout_and_retry(
            monitor_fn,
            timeout_secs=timeout_secs,
            max_retries=max_retries,
            retry_delay_secs=retry_delay_secs,
        ):
            on_fail()
            break
        else:
            await asyncio.sleep(period_secs)


async def _check_with_timeout_and_retry(
    monitor_fn: Callable[[], Awaitable[bool]],
    timeout_secs: int = 60,
    max_retries: int = 3,
    retry_delay_secs: int = 2,
) -> Awaitable[bool]:
    with_timeout = predicate_with_timeout(monitor_fn, timeout_secs=timeout_secs)
    with_retry = retry_predicate(with_timeout, max_retries=max_retries, delay_secs=retry_delay_secs)
    return await with_retry()


async def test_predict(async_predict_fn: Callable, logger: Logger) -> bool:
    try:
        model_input = {
            "prompt": "test",
            "max_tokens": 1,
        }
        fastapi_request = StarletteRequest(
            scope={
                "type": "http",
                "method": "POST",
            }
        )
        fastapi_request.is_disconnected = lambda: False
        await async_predict_fn(model_input, fastapi_request)
        return True
    except Exception as e:
        logger.info(f"Error during prediction test: {e}")
        return False


async def start_monitor(
    briton_interactor: BritonInteractor, async_predict_fn: Callable, logger: Logger
):
    monitor_settings = briton_interactor.monitor_settings()
    if not monitor_settings.start_monitor_thread:
        return

    def on_fail():
        logger.error("Truss server is stuck, exiting truss server")
        kill_current_process_and_children()

    async def monitor_fn():
        try:
            return await test_predict(async_predict_fn, logger)
        except Exception as e:
            return False

    asyncio.create_task(
        monitor(
            monitor_fn,
            on_fail,
            monitor_settings.should_continue,
            monitor_settings.period_secs,
            monitor_settings.max_retries,
            monitor_settings.timeout_secs,
            monitor_settings.retry_delay_secs,
        )
    )



# Briton startup
# Briton inference request creation

import abc
import asyncio
import atexit
import os
import signal
import subprocess
import threading
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional

import grpc
from huggingface_hub import snapshot_download
from transformers import AutoTokenizer, PreTrainedTokenizer, PreTrainedTokenizerFast

from briton.constants import FSM_CACHE_DIR
from briton.flags import TRUSS_DEVELOPMENT_MODE
from briton.fs import is_local_path
from briton.network import is_port_available
from briton.proto import BritonStub
from briton.tokenizer import serialize_added_tokens_to_config

BRITON_CONFIG_FILENAME = "briton_config.pbtxt"
BRITON_STARTUP_CHECK_FREQUENCY_SECS = 1
BRITON_STARTUP_LOG_FREQUENCY_SECS = 10  # Print fewer log lines
BRITON_MONITOR_FREQUENCY_SECS = 1


@dataclass
class MonitorSettings:
    start_monitor_thread: bool = True
    period_secs: int = 30
    max_retries: int = 3
    timeout_secs: int = 60
    retry_delay_secs: int = 1
    should_continue: Optional[Callable[[], bool]] = None


class EosRespondingStub:
    """Only returns responses based on given input ids."""

    def __init__(self, eos_token_id: int, eos_token: str):
        self._eos_token_id = eos_token_id
        self._eos_token = eos_token

    def Infer(self, request):
        responses = []
        responses.append(
            MockDelta(
                [self._eos_token_id],
                self._eos_token,
            )
        )
        return MockInferAsyncIterator(responses)


class FailingStub:
    def Infer(self, request):
        raise RuntimeError("Simulated error")


class BritonInteractor(abc.ABC):
    @abc.abstractmethod
    def hf_download(self, repo_id: str, local_dir: Path, hf_token: Optional[str] = None):
        pass

    @abc.abstractmethod
    def load(
        self,
        model_name: str,
        engine_path: str,
        hf_tokenizer: str,
        work_dir: Path,
        fsm_cache_dir: str,
        kv_cache_free_gpu_mem_fraction: float,
        port: int,
        added_tokens: list,
        max_num_tokens: Optional[int],
        enable_chunked_context: bool = False,
        hf_token: Optional[str] = None,
        tp_count: Optional[int] = 1,
        batch_scheduler_policy: Optional[int] = None,
    ):
        pass

    # TODO: Get typing on this right.
    @abc.abstractmethod
    def create_grpc_stub(self, port: int) -> Any:
        pass

    @abc.abstractmethod
    def auto_tokenizer_from_pretrained(
        self, hf_model_repo: str, hf_token: Optional[str] = None
    ) -> PreTrainedTokenizerFast | PreTrainedTokenizer:
        pass

    @abc.abstractmethod
    def fsm_cache_dir(self) -> Path:
        pass

    @abc.abstractmethod
    def monitor_settings(self) -> MonitorSettings:
        pass


class BritonInteractorImpl(BritonInteractor):
    def hf_download(self, repo_id: str, local_dir: Path, hf_token: Optional[str] = None):
        snapshot_download(repo_id, local_dir=local_dir, token=hf_token)

    def load(self, *args, **kwargs):
        return load_briton(*args, **kwargs)

    def create_grpc_stub(self, port: int) -> BritonStub:
        channel = grpc.aio.insecure_channel(f"localhost:{port}")
        return BritonStub(channel)

    def auto_tokenizer_from_pretrained(self, hf_model_repo: str, hf_token: Optional[str] = None):
        return AutoTokenizer.from_pretrained(hf_model_repo, token=hf_token, trust_remote_code=True)

    def fsm_cache_dir(self) -> Path:
        return Path(FSM_CACHE_DIR)

    def monitor_settings(self) -> MonitorSettings:
        return MonitorSettings()


class MockDelta:
    def __init__(self, output_ids, output_text=None, output_states=None):
        self.output_ids = output_ids
        self.output_text = output_text
        # TODO(pankaj) Set these values
        self.output_states = output_states or []


class MockInferAsyncIterator:
    def __init__(self, responses):
        self._responses = responses
        self._index = 0

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self._index < len(self._responses):
            await asyncio.sleep(0.01)
            response = self._responses[self._index]
            self._index += 1
            return response
        else:
            raise StopAsyncIteration


class EosRespondingBritonInteractor(BritonInteractor):
    def __init__(self):
        self._tokenizer = AutoTokenizer.from_pretrained("TinyLlama/TinyLlama-1.1B-Chat-v1.0")

    def hf_download(self, repo_id: str, local_dir: Path, hf_token: Optional[str] = None):
        pass

    def load(self, *args, **kwargs):
        pass

    def create_grpc_stub(self, port: int) -> EosRespondingStub:
        assert self._tokenizer.eos_token_id is not None
        return EosRespondingStub(self._tokenizer.eos_token_id, self._tokenizer.eos_token)

    def auto_tokenizer_from_pretrained(self, hf_model_repo: str, hf_token: Optional[str] = None):
        return self._tokenizer

    def fsm_cache_dir(self) -> Path:
        return Path(FSM_CACHE_DIR)

    def monitor_settings(self) -> MonitorSettings:
        return MonitorSettings()


class FailingBritonInteractor(BritonInteractor):
    def __init__(self):
        self._tokenizer = AutoTokenizer.from_pretrained("TinyLlama/TinyLlama-1.1B-Chat-v1.0")

    def hf_download(self, repo_id: str, local_dir: Path, hf_token: Optional[str] = None):
        pass

    def load(self, *args, **kwargs):
        pass

    def create_grpc_stub(self, port: int) -> FailingStub:
        return FailingStub()

    def auto_tokenizer_from_pretrained(self, hf_model_repo: str, hf_token: Optional[str] = None):
        return self._tokenizer

    def fsm_cache_dir(self) -> Path:
        return Path(FSM_CACHE_DIR)

    def monitor_settings(self) -> MonitorSettings:
        return MonitorSettings(
            period_secs=1,
            timeout_secs=1,
        )


def load_briton(
    model_name: str,
    engine_path: str,
    hf_tokenizer: str,
    work_dir: Path,
    fsm_cache_dir: str,
    kv_cache_free_gpu_mem_fraction: float,
    port: int,
    added_tokens: list,
    max_num_tokens: Optional[int],
    enable_chunked_context: bool = False,
    hf_token: Optional[str] = None,
    tp_count: Optional[int] = 1,
    batch_scheduler_policy: Optional[int] = None,
):
    """Starts a Briton server for a given model type.

    TODO: Document the parameters.
    """
    if TRUSS_DEVELOPMENT_MODE:
        # Loading models (via Briton) can be slow. In development mode we reuse existing
        # Briton servers. If the port is occupied we assume one is running and we don't
        # start a new one.
        if not is_port_available(port):
            return

    # TODO(pankaj) Use this after debugging an issue we ran into with this.
    # Pass tokenizer file to Briton for the rust tokenizer.
    if is_local_path(hf_tokenizer):
        hf_tokenizer = str(Path(hf_tokenizer) / "tokenizer.json")

    config_str = f"""
        engine_path: "{engine_path}"
        hf_tokenizer: "{hf_tokenizer}"
        kv_cache_free_gpu_mem_fraction: {kv_cache_free_gpu_mem_fraction}
        enable_kv_cache_reuse: true
        enable_chunked_context: {enable_chunked_context}
        port: {port}
        fsm_cache_dir: "{fsm_cache_dir}"
    """
    if max_num_tokens is not None:
        config_str += f"\nmax_num_tokens: {max_num_tokens}"
    if batch_scheduler_policy is not None:
        config_str += f"\nbatch_scheduler_policy: {batch_scheduler_policy}"
    # Pass added tokens to Briton for the rust tokenizer.
    if len(added_tokens) > 0:
        config_str += "\n" + "\n".join(serialize_added_tokens_to_config(added_tokens))

    work_dir.mkdir(parents=True, exist_ok=True)
    config_pbtxt_path = (work_dir / BRITON_CONFIG_FILENAME).resolve()
    config_pbtxt_path.write_text(config_str)
    briton_env = os.environ.copy()
    if hf_token is not None:
        briton_env["HF_ACCESS_TOKEN"] = hf_token
        briton_env["HF_TOKEN"] = hf_token
    briton_process = _start_briton(config_pbtxt_path, tp_count, briton_env)
    # It can sometimes take a couple of minutes for Briton to start
    # This counter prevents up to 810 unnecessary log lines
    counter = 0
    while is_port_available(port):
        if counter % BRITON_STARTUP_LOG_FREQUENCY_SECS == 0:
            print(f"Waiting for Briton server for {model_name} to start")
        counter += 1
        time.sleep(BRITON_STARTUP_CHECK_FREQUENCY_SECS)

    briton_monitor_thread = threading.Thread(target=_briton_monitor, args=(briton_process,))
    briton_monitor_thread.daemon = True
    briton_monitor_thread.start()


def _briton_monitor(briton_process):
    while True:
        if briton_process.poll() is not None:
            print(
                f"Briton process has exited with code {briton_process.returncode}, exiting truss server"
            )
            pid = os.getpid()
            os.kill(pid, signal.SIGKILL)
        time.sleep(BRITON_MONITOR_FREQUENCY_SECS)


def _start_briton(config_pbtxt_path, tp_count, briton_env):
    if tp_count is None or tp_count == 1:
        briton_process = subprocess.Popen(
            ["Briton", "--config", str(config_pbtxt_path)], env=briton_env
        )
    else:
        briton_process = subprocess.Popen(
            [
                "mpirun",
                "--allow-run-as-root",
                "-n",
                f"{tp_count}",
                "Briton",
                "--config",
                str(config_pbtxt_path),
            ],
            env=briton_env,
        )
    atexit.register(_cleanup_subprocess, briton_process)
    return briton_process


def _cleanup_subprocess(proc):
    print("Cleaning up: terminating the subprocess")
    try:
        os.killpg(os.getpgid(proc.pid), signal.SIGTERM)  # Terminate the subprocess group
    except Exception as e:
        print(f"Failed to terminate subprocess: {e}")



from typing import Any, AsyncGenerator, Awaitable, Callable, Dict, List, Optional, Union

from fastapi import HTTPException
from openai.types.chat import ChatCompletion

from briton.backend.utils import collect_text
from briton.constants import BRITON_DEFAULT_MAX_TOKENS
from briton.error_handling import grpc_error_handling
from briton.openai import create_completion, create_completion_chunks
from briton.proto import InferenceRequest
from briton.schema import ModelInput


async def model_input_to_briton_request(
    request_id: int,
    model_input: ModelInput,
    input_ids: List[int],
    tokenizer_eos_token_id: int,
    tokenizer_pad_token_id: int,
    add_schema_to_cache: Callable[[Dict[str, Any]], Awaitable[str]],
    default_max_tokens: int,
    max_seq_len: int,
) -> InferenceRequest:
    """
    Convert ModelInput to Briton inference request.

    :param model_input: model input
    :return: InferRequest
    """
    request = InferenceRequest(
        request_id=request_id,
        input_ids=input_ids,
    )
    num_input_ids = len(input_ids)

    # end_id, pad_id
    end_id = model_input.end_id or tokenizer_eos_token_id
    if end_id is not None:
        request.end_id = end_id
    pad_id = model_input.pad_id or tokenizer_pad_token_id
    if pad_id is not None:
        request.pad_id = pad_id

    schema = model_input.output_json_schema
    if schema is not None:
        request.output_schema_hash = await add_schema_to_cache(schema)
        force_tools = model_input.force_tools
        if force_tools is not None:
            request.force_tools = force_tools

    # max tokens
    request.request_output_len = calc_request_output_len(
        model_input.max_tokens,
        num_input_ids,
        default_max_tokens,
        max_seq_len,
    )

    # beam width
    if model_input.beam_width is not None:
        request.beam_width = model_input.beam_width

    # repetition penalty
    if model_input.repetition_penalty is not None:
        request.repetition_penalty = model_input.repetition_penalty

    # presence_penalty
    if model_input.presence_penalty is not None:
        request.presence_penalty = model_input.presence_penalty

    # length penalty
    if model_input.length_penalty is not None:
        request.len_penalty = model_input.length_penalty

    # temperature
    if model_input.temperature is not None:
        request.temperature = model_input.temperature

    # length_penalty
    if model_input.length_penalty is not None:
        request.len_penalty = model_input.length_penalty

    # frequency_penalty
    if model_input.frequency_penalty is not None:
        # TODO(pankaj) Implement this in Briton
        # request.frequency_penalty = model_input.frequency_penalty
        pass

    # top_k
    if model_input.top_k is not None:
        request.runtime_top_k = model_input.top_k

    # top_p
    if model_input.top_p is not None:
        request.runtime_top_p = model_input.top_p

    # random seed
    # Note that model_input always has a seed. If seed is not supplied in input
    # then a random value is picked.
    request.random_seed = model_input.seed

    # stop words
    if model_input.stop is not None:
        request.stop_words.extend(model_input.stop)

    # bad words
    if model_input.bad_words is not None:
        request.bad_words.extend(model_input.bad_words)

    return request


def calc_request_output_len(
    model_input_max_tokens: Optional[int],
    num_input_ids: int,
    default_max_tokens: Optional[int],
    max_seq_len: int,
) -> int:
    if num_input_ids > max_seq_len:
        raise HTTPException(
            status_code=400,
            detail=f"Number of input tokens {num_input_ids} exceeds max sequence length {max_seq_len}",
        )
    model_max_allowed_tokens = max_seq_len - num_input_ids
    max_tokens = model_input_max_tokens or default_max_tokens or BRITON_DEFAULT_MAX_TOKENS
    return min(model_max_allowed_tokens, max_tokens)


async def openai_spec_response(
    resp_iter: AsyncGenerator[Any, None],
    request_id: str,
    num_input_ids: int,
    streaming: bool,
    eos_token: str,
    tool_call_token: Optional[str],
    model_name: str,
    include_stream_usage: bool,
    stop_words: Optional[List[str]],
    skip_special_tokens: Optional[List[str]],
) -> Union[AsyncGenerator[str, None], ChatCompletion]:
    with grpc_error_handling():
        if streaming:
            completion_tokens = 0
            # Advance the iterator to get the first chunk, to allow any
            # validation error to be thrown. Any errors will be handled
            # gy grpc_error_handling and converted to HTTPException.
            # `first_chunk` is not discarded, it is used below.
            async for first_chunk in resp_iter:
                break

            async def first_chunk_and_generate_after():
                nonlocal completion_tokens
                completion_tokens += len(first_chunk.output_ids)
                yield first_chunk.output_text
                async for chunk in resp_iter:
                    completion_tokens += len(chunk.output_ids)
                    yield chunk.output_text

            def completion_tokens_fn():
                nonlocal completion_tokens
                return completion_tokens

            return create_completion_chunks(
                req_id=request_id,
                model=model_name,
                input_text=first_chunk_and_generate_after(),
                eos_token=eos_token,
                tool_token=tool_call_token,
                prompt_tokens=(num_input_ids if include_stream_usage else None),
                completion_tokens_fn=(completion_tokens_fn if include_stream_usage else None),
                stop_words=stop_words,
                skip_special_tokens=skip_special_tokens,
            )
        else:
            full_text, completion_tokens = await collect_text(resp_iter)
            return create_completion(
                req_id=request_id,
                model=model_name,
                input_text=full_text,
                eos_token=eos_token,
                tool_token=tool_call_token,
                prompt_tokens=num_input_ids,
                completion_tokens=completion_tokens,
                stop_words=stop_words,
                skip_special_tokens=skip_special_tokens,
            )



from typing import List, Optional

from briton.backend.backend_types import RequestDetails
from briton.backend.default_backend import DefaultBackend
from briton.backend.utils import collect_text
from briton.data_structures import or_true
from briton.error_handling import grpc_error_handling
from briton.openai import remove_suffix_from_text
from briton.schema import ModelInput


class LegacyBackend(DefaultBackend):
    def __init__(self):
        super().__init__()

    # Overrides parent
    async def accepts_request(self, model_input: ModelInput) -> Optional[RequestDetails]:
        if self._config is None or self._tokenizer is None:
            return None

        if self._config.is_openai_compatible:
            return None

        input_ids = model_input.input_ids(self._tokenizer)
        return RequestDetails(input_ids=input_ids)

    # Overrides parent
    async def generate_response(
        self,
        resp_iter,
        model_input: ModelInput,
        request_id: str,
        input_ids: List[int],
        eos_token: str,
        tool_call_token: str,
    ):
        streaming = or_true(model_input.stream)
        with grpc_error_handling():
            if streaming:
                # Advance the iterator to get the first chunk, to allow any
                # validation error to be thrown. Any errors will be handled
                # gy grpc_error_handling and converted to HTTPException.
                # `first_chunk` is not discarded, it is used below.
                async for first_chunk in resp_iter:
                    break

                async def generate_processed_after_first_chunk():
                    output_text = first_chunk.output_text
                    # Try to remove tool call token from first chunk only
                    if tool_call_token is not None:
                        output_text = output_text.removeprefix(tool_call_token)
                    yield remove_suffix_from_text(
                        output_text,
                        eos_token,
                        model_input.stop,
                        model_input.skip_special_tokens,
                    )

                    async for chunk in resp_iter:
                        output_text = chunk.output_text
                        yield remove_suffix_from_text(
                            output_text,
                            eos_token,
                            model_input.stop,
                            model_input.skip_special_tokens,
                        )

                return generate_processed_after_first_chunk()
            else:
                full_text, _ = await collect_text(resp_iter)
                full_text = remove_suffix_from_text(
                    full_text,
                    eos_token,
                    model_input.stop,
                    model_input.skip_special_tokens,
                )
                if tool_call_token is not None:
                    full_text = full_text.removeprefix(tool_call_token)
                return full_text



import random
from functools import wraps
from typing import Any, Dict, List, Literal, Optional, Union, cast

from fastapi import HTTPException
from jinja2.exceptions import TemplateError
from pydantic import BaseModel, Field, ValidationError, model_validator
from transformers import PreTrainedTokenizerFast

from briton.constants import UINT32_MAX


def handle_tokenizer_exceptions(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (ValueError, TemplateError) as e:
            raise HTTPException(status_code=400, detail=str(e))

    return wrapper


class ModelInput(BaseModel):
    """This class mirrors the `CompletionCreateParamsBase` in the `openai-python` repository.

    However, that class is a TypedDict rather than a pydantic model, so we redefine it here
    to take advantage of pydantic's validation features. In addition, we define helper methods
    to get the formatted prompt, tools to use, and response format to adhere to.

    Unsupported parameters:
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-store
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-metadata
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-frequency_penalty
      - Frequency penalty is not currently passed through to briton
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-logit_bias
      - User provided logit biasing is not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-logprobs
      - Returning log probabilities is not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-top_logprobs
      - Returning log probabilities is not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-n
      - Multiple generation is not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-service_tier
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-user
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-function_call
      - Deprecated
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-functions
      - Deprecated
    """

    class Tool(BaseModel):
        """An element in the top level `tools` field."""

        class Function(BaseModel):
            name: str
            description: Optional[str] = Field(None)
            parameters: Dict[str, Any]  # `parameters` holds the json schema
            return_: Optional[Dict[str, Any]] = Field(None, alias="return")

            @model_validator(mode="after")
            def definitions_valid(cls, values):
                if "definitions" in values.parameters and "$defs" in values.parameters:
                    raise ValueError(
                        "Both pydantic v1 and v2 definitions found; please check schema."
                    )
                return values

            @property
            def parameters_without_definitions(self) -> Dict[str, Any]:
                parameters = self.parameters.copy()
                for keyword in ["definitions", "$defs"]:
                    parameters.pop(keyword, None)
                return parameters

            @property
            def definitions(self) -> Optional[tuple[Dict[str, Any], str]]:
                for keyword in ["definitions", "$defs"]:
                    if keyword in self.parameters:
                        return self.parameters[keyword], keyword
                return None

            @property
            def json_schema(self) -> Dict[str, Any]:
                return {
                    "type": "object",
                    "properties": {
                        "name": {"const": self.name},
                        "parameters": self.parameters_without_definitions,
                    },
                    "required": ["name", "parameters"],
                }

        type: Literal["function"]
        function: Function

    class ToolChoice(BaseModel):
        """The top level `tool_choice` field."""

        class FunctionChoice(BaseModel):
            name: str

        type: Literal["function"]
        function: FunctionChoice

    class SchemaResponseFormat(BaseModel):
        """The top level `response_format` field."""

        class JsonSchema(BaseModel):
            """`schema_` holds the actual json schema"""

            schema_: Dict[str, Any] = Field(..., alias="schema")

        type: Literal["json_schema"]
        json_schema: JsonSchema

    class JsonResponseFormat(BaseModel):
        type: Literal["json_object"]

    class TextResponseFormat(BaseModel):
        type: Literal["text"]

    class StreamOptions(BaseModel):
        """The top level `stream_options` field."""

        include_usage: bool

    model: Optional[str] = Field("")

    # TODO: Define `Message` objects to mirror `ChatCompletionMessageParam` for validation
    messages: Optional[List[Dict[str, Any]]] = Field(None)
    prompt_: Optional[str] = Field(None, min_length=1, alias="prompt")

    max_tokens_: Optional[int] = Field(None, alias="max_tokens")
    max_completion_tokens: Optional[int] = Field(None)

    stream: Optional[bool] = Field(None)
    stream_options: Optional[StreamOptions] = Field(None)

    seed_: Optional[int] = Field(None, alias="seed")
    random_seed: Optional[int] = Field(None)

    # TODO(bdubayah) : Wire this up to briton
    frequency_penalty: Optional[float] = Field(0)
    presence_penalty: Optional[float] = Field(0)
    length_penalty: Optional[float] = Field(None)

    # Not part of openai spec but supported by briton
    repetition_penalty: Optional[float] = Field(None)

    temperature: Optional[float] = Field(1.0)
    top_p_: Optional[float] = Field(1.0, alias="top_p")
    runtime_top_p: Optional[float] = Field(None)
    top_k_: Optional[int] = Field(50, alias="top_k")
    runtime_top_k: Optional[int] = Field(None)
    stop_: Optional[Union[str, List[str]]] = Field(None, alias="stop")
    bad_words_: Optional[Union[str, List[str]]] = Field(None, alias="bad_words")
    skip_special_tokens: Optional[List[str]] = Field(None)

    response_format: Optional[
        Union[SchemaResponseFormat, JsonResponseFormat, TextResponseFormat]
    ] = Field(None)
    tools_: Optional[List[Tool]] = Field(None, alias="tools")
    tool_choice: Optional[Union[Literal["none", "required", "auto"], ToolChoice]] = Field(None)
    parallel_tool_calls: Optional[bool] = Field(True)

    beam_width: Optional[Literal[1]] = Field(None)

    end_id: Optional[int] = Field(None)
    pad_id: Optional[int] = Field(None)

    _generated_seed: Optional[int] = None

    # WiM fields
    margins_prompt: Optional[str] = Field(None)
    margins_stop_sequences: Optional[List[str]] = Field(["NO#"])
    max_chunk_size: Optional[int] = Field(4096)

    @model_validator(mode="after")
    def top_p_valid(cls, values):
        if values.runtime_top_p is None:
            values.runtime_top_p = values.top_p_
        if values.top_p < 0 or values.top_p > 1:
            raise ValueError("`top_p` must be between 0 and 1.")
        return values

    @model_validator(mode="after")
    def top_k_valid(cls, values):
        if values.runtime_top_k is None:
            values.runtime_top_k = values.top_k_
        if values.top_k < 0:
            raise ValueError("`top_k` must be greater than or equal to 0.")
        if values.top_k is not None and values.top_k > UINT32_MAX:
            raise ValueError(f"`top_k` must be less than or equal to {UINT32_MAX}.")
        return values

    @model_validator(mode="after")
    def messages_not_empty(cls, values):
        messages = values.messages
        if messages is not None and len(messages) == 0:
            raise ValueError("`messages` cannot be empty.")
        return values

    @model_validator(mode="after")
    def oneof_messages_and_prompt_set(cls, values):
        prompt = values.prompt_
        messages = values.messages
        if prompt is None and messages is None:
            raise ValueError("One of `prompt` or `messages` must be specified.")
        if prompt is not None and messages is not None:
            raise ValueError("Only one of `prompt` and `messages` can be specified.")
        return values

    @model_validator(mode="after")
    def max_tokens_and_max_completion_tokens_not_set(cls, values):
        max_tokens = values.max_tokens_
        max_completion_tokens = values.max_completion_tokens
        if max_tokens is not None and max_completion_tokens is not None:
            raise ValueError(
                "Only one of `max_tokens` and `max_completion_tokens` can be specified."
            )
        return values

    @model_validator(mode="after")
    def tools_valid(cls, values):
        tools = values.tools_
        tool_choice = values.tool_choice
        if tools is not None and tool_choice is None:
            values.tool_choice = "auto"
        if tools is not None and len(tools) == 0 and tool_choice != "none":
            raise ValueError("`tools` cannot be empty.")
        if isinstance(tool_choice, cls.ToolChoice) and tool_choice.function.name not in [
            tool.function.name for tool in tools
        ]:
            raise ValueError("`tool_choice` not in `tools`.")
        return values

    @model_validator(mode="after")
    def tools_not_used_with_prompt(cls, values):
        prompt = values.prompt_
        tool_choice = values.tool_choice
        if prompt is not None and tool_choice is not None and tool_choice != "none":
            raise ValueError("`tool_choice` cannot be used with `prompt`.")
        return values

    @model_validator(mode="after")
    def tools_not_used_with_response_format(cls, values):
        response_format = values.response_format
        tool_choice = values.tool_choice
        if response_format is not None and tool_choice is not None and tool_choice != "none":
            raise ValueError("`tools` cannot be used with `response_format`.")
        return values

    @model_validator(mode="after")
    def temperature_ge_0(cls, values):
        if values.temperature < 0:
            raise ValueError("`temperature` must be >= 0.")
        return values

    @model_validator(mode="after")
    def adjust_temperature_for_greedy_decoding(cls, values):
        if values.temperature == 0:
            values.temperature = 0.01
            values.top_k_ = 0
            values.runtime_top_k = 0
            values.top_p_ = 0.0
            values.runtime_top_p = 0.0
        return values

    @property
    def force_tools(self) -> Optional[bool]:
        if self.tool_choice is not None and self.tool_choice != "none":
            return self.tool_choice == "required" or isinstance(self.tool_choice, self.ToolChoice)
        return None

    @property
    def tools(self) -> Optional[List[Tool]]:
        """Returns the tools to use, dependent on tool_choice."""
        if self.tool_choice is not None and self.tool_choice != "none":
            if self.tools_ is None:
                raise ValueError("`tools` must be specified with `tool_choice`.")
            if isinstance(self.tool_choice, self.ToolChoice):
                return [
                    tool
                    for tool in self.tools_
                    if tool.function.name == self.tool_choice.function.name
                ]
            return self.tools_
        return None

    @property
    def _tool_dicts(self) -> Optional[List[Dict[str, Any]]]:
        """Convenience property to get all tools as plain dicts."""
        return (
            [tool.model_dump(by_alias=True) for tool in self.tools]
            if self.tools is not None
            else None
        )

    @staticmethod
    def _create_tools_schema(
        tools: List[Tool], parallel_tool_calls: bool
    ) -> Optional[Dict[str, Any]]:
        schema = {
            "type": "array",
            "items": {"anyOf": [tool.function.json_schema for tool in tools]},
            "minItems": 1,
        }
        if not parallel_tool_calls:
            schema["maxItems"] = 1

        # This block handles moving nested definitions to the top level of the schema.
        # Because we support parallel tool calls via `anyOf`, we need to combine all
        # definitions across all tools into a single object.
        # We also have to handle the different keywords used for definitions between
        # pydantic v1 and v2.
        combined_definitions = {}
        combined_keyword = None
        for tool in tools:
            if tool.function.definitions is not None:
                definitions, keyword = tool.function.definitions
                if combined_keyword is not None and keyword != combined_keyword:
                    raise ValueError(
                        "Both pydantic v1 and v2 definitions found; please check schema."
                    )
                combined_keyword = keyword
                for name, definition in definitions.items():
                    if name in combined_definitions and combined_definitions[name] != definition:
                        raise ValueError(f"Conflicting definitions for '{name}' found.")
                    combined_definitions[name] = definition
        if combined_keyword is not None and len(combined_definitions) > 0:
            schema[combined_keyword] = combined_definitions

        return schema

    @property
    def output_json_schema(self) -> Optional[Dict[str, Any]]:
        """Creates the output json schema based on the response format or tools."""
        if isinstance(self.response_format, self.SchemaResponseFormat):
            return self.response_format.json_schema.schema_
        if isinstance(self.response_format, self.TextResponseFormat):
            return None
        if isinstance(self.response_format, self.JsonResponseFormat):
            return {}
        if self.tools is not None:
            return self._create_tools_schema(self.tools, self.parallel_tool_calls != False)
        return None

    @property
    def max_tokens(self) -> Optional[int]:
        """`max_tokens` was deprecated in favor of `max_completion_tokens`"""
        return self.max_tokens_ if self.max_tokens_ is not None else self.max_completion_tokens

    @property
    def top_p(self) -> Optional[float]:
        """`top_p` was previously named `runtime_top_p` in briton"""
        return self.runtime_top_p if self.runtime_top_p is not None else self.top_p_

    @property
    def top_k(self) -> Optional[int]:
        """`top_k` was previously named `runtime_top_k` in briton"""
        return self.runtime_top_k if self.runtime_top_k is not None else self.top_k_

    @property
    def seed(self) -> int:
        """`seed` was previously named `random_seed` in briton"""
        if self.seed_ is not None:
            return self.seed_
        if self.random_seed is not None:
            return self.random_seed
        if self._generated_seed is None:
            self._generated_seed = random.randint(-(2**63), 2**63 - 1)
        return self._generated_seed

    @property
    def include_stream_usage(self) -> bool:
        return self.stream_options is not None and self.stream_options.include_usage

    @property
    def combined_messages(self) -> List[Dict[str, Any]]:
        """Combine consecutive user messages into a single message to avoid chat template errors."""
        if not self.messages:
            return []

        combined_messages = []
        current_user_message = None

        for message in self.messages:
            if message["role"] == "user":
                if current_user_message:
                    # OpenAI combines user messages with a space and newline
                    current_user_message["content"] += f" \n{message['content']}"
                else:
                    current_user_message = message.copy()
            else:
                if current_user_message:
                    combined_messages.append(current_user_message)
                    current_user_message = None
                combined_messages.append(message)

        if current_user_message:
            combined_messages.append(current_user_message)

        return combined_messages

    @property
    def stop(self) -> Optional[List[str]]:
        if self.stop_ is None:
            return None
        if isinstance(self.stop_, str):
            return [self.stop_]
        return self.stop_

    @property
    def bad_words(self) -> Optional[List[str]]:
        if self.bad_words_ is None:
            return None
        if isinstance(self.bad_words_, str):
            return [self.bad_words_]
        return self.bad_words_

    def prompt(self, tokenizer: PreTrainedTokenizerFast) -> str:
        """Calculate text prompt from model_input.

        Prompt may be supplied in the input as such or as messages. If messages
        are supplied, they are used to generate the prompt using chat template.
        """
        if self.prompt_ is None:
            if self.messages is None:
                raise ValueError("`messages` must be specified.")
            return cast(
                str,
                tokenizer.apply_chat_template(
                    conversation=self.combined_messages,
                    tools=self._tool_dicts,
                    tokenize=False,
                    add_generation_prompt=True,
                ),
            )
        return self.prompt_

    @handle_tokenizer_exceptions
    def input_ids(self, tokenizer: PreTrainedTokenizerFast) -> List[int]:
        """Get the input ids from the prompt."""
        return tokenizer.encode(self.prompt(tokenizer), add_special_tokens=False)


class ModelInputWithoutTools(ModelInput):
    """Input class for models that don't support function calling."""

    @model_validator(mode="after")
    def tools_not_used(cls, values):
        if values.tools is not None:
            raise ValueError("This model does not support `tools`.")
        return values


def get_prompt(model_input: ModelInput, tokenizer: PreTrainedTokenizerFast) -> str:
    try:
        return model_input.prompt(tokenizer)
    except (ValueError, TemplateError) as e:
        raise HTTPException(status_code=400, detail=str(e))


def validate_model_input(model_input: Dict[str, Any], supports_tools: bool = True) -> ModelInput:
    try:
        if supports_tools:
            return ModelInput(**model_input)
        else:
            return ModelInputWithoutTools(**model_input)
    except (ValueError, ValidationError) as e:
        raise HTTPException(status_code=400, detail=str(e))


def update_raw_model_input(model_input: Dict[str, Any], validated_input: ModelInput) -> None:
    """Updates the `model_input` dictionary with values from the `validated_input` object.

    This is need in cases where 1) the field names different between the spec and briton
    and 2) where we override values (in the case of `seed` or `temperature`).

    Parameters:
    - model_input (Dict[str, Any]): The raw model input.
    - validated_input (ModelInput): The validated model input.
    """
    if validated_input.max_tokens is not None:
        model_input["max_tokens"] = validated_input.max_tokens
    if validated_input.top_k is not None:
        model_input["runtime_top_k"] = validated_input.top_k
    if validated_input.top_p is not None:
        model_input["runtime_top_p"] = validated_input.top_p
    if validated_input.temperature is not None:
        model_input["temperature"] = validated_input.temperature
    if validated_input.stop is not None:
        model_input["stop_words"] = validated_input.stop
    if validated_input.bad_words is not None:
        model_input["bad_words"] = validated_input.bad_words
    model_input["random_seed"] = validated_input.seed



import json
import warnings
from enum import Enum
from typing import Optional

from huggingface_hub.errors import HFValidationError
from huggingface_hub.utils import validate_repo_id
from pydantic import BaseModel, PydanticDeprecatedSince20, validator

# Suppress Pydantic V1 warnings, because we have to use it for backwards compat.
warnings.filterwarnings("ignore", category=PydanticDeprecatedSince20)


class TrussTRTLLMModel(str, Enum):
    LLAMA = "llama"
    MISTRAL = "mistral"
    DEEPSEEK = "deepseek"
    WHISPER = "whisper"
    QWEN = "qwen"
    PALMYRA = "palmyra"


class TrussTRTLLMQuantizationType(str, Enum):
    NO_QUANT = "no_quant"
    WEIGHTS_ONLY_INT8 = "weights_int8"
    WEIGHTS_KV_INT8 = "weights_kv_int8"
    WEIGHTS_ONLY_INT4 = "weights_int4"
    WEIGHTS_INT4_KV_INT8 = "weights_int4_kv_int8"
    SMOOTH_QUANT = "smooth_quant"
    FP8 = "fp8"
    FP8_KV = "fp8_kv"


class TrussTRTLLMPluginConfiguration(BaseModel):
    paged_kv_cache: bool = True
    gemm_plugin: str = "auto"
    use_paged_context_fmha: bool = False
    use_fp8_context_fmha: bool = False


class CheckpointSource(str, Enum):
    HF: str = "HF"
    GCS: str = "GCS"
    LOCAL: str = "LOCAL"
    # REMOTE_URL is useful when the checkpoint lives on remote storage accessible via HTTP (e.g a presigned URL)
    REMOTE_URL: str = "REMOTE_URL"


class CheckpointRepository(BaseModel):
    source: CheckpointSource
    repo: str


class TrussTRTLLMBatchSchedulerPolicy(str, Enum):
    MAX_UTILIZATION = "max_utilization"
    GUARANTEED_NO_EVICT = "guaranteed_no_evict"


class TrussSpecDecMode(str, Enum):
    DRAFT_EXTERNAL: str = "DRAFT_TOKENS_EXTERNAL"


class TrussTRTLLMBuildConfiguration(BaseModel):
    base_model: TrussTRTLLMModel
    max_seq_len: int
    max_batch_size: Optional[int] = 256
    max_num_tokens: Optional[int] = 8192
    max_beam_width: int = 1
    max_prompt_embedding_table_size: int = 0
    checkpoint_repository: CheckpointRepository
    gather_all_token_logits: bool = False
    strongly_typed: bool = False
    quantization_type: TrussTRTLLMQuantizationType = TrussTRTLLMQuantizationType.NO_QUANT
    tensor_parallel_count: int = 1
    pipeline_parallel_count: int = 1
    plugin_configuration: TrussTRTLLMPluginConfiguration = TrussTRTLLMPluginConfiguration()
    num_builder_gpus: Optional[int] = None
    speculative_decoding_mode: Optional[TrussSpecDecMode] = None
    max_draft_len: Optional[int] = None

    @validator("max_beam_width")
    def check_max_beam_width(cls, v: int):
        if isinstance(v, int):
            if v != 1:
                raise ValueError("max_beam_width greater than 1 is not currently supported")
        return v


class TrussTRTLLMRuntimeConfiguration(BaseModel):
    kv_cache_free_gpu_mem_fraction: float = 0.9
    enable_chunked_context: bool = False
    num_draft_tokens: Optional[int] = None
    batch_scheduler_policy: TrussTRTLLMBatchSchedulerPolicy = (
        TrussTRTLLMBatchSchedulerPolicy.GUARANTEED_NO_EVICT
    )
    request_default_max_tokens: Optional[int] = None


class TRTLLMConfiguration(BaseModel):
    runtime: TrussTRTLLMRuntimeConfiguration = TrussTRTLLMRuntimeConfiguration()
    build: TrussTRTLLMBuildConfiguration

    def __init__(self, **data):
        super().__init__(**data)
        self._validate_kv_cache_flags()
        if self.build.checkpoint_repository.source == CheckpointSource.HF:
            self._validate_hf_repo_id()

    def _validate_kv_cache_flags(self):
        if self.build is None:
            return self
        if not self.build.plugin_configuration.paged_kv_cache and (
            self.build.plugin_configuration.use_paged_context_fmha
            or self.build.plugin_configuration.use_fp8_context_fmha
        ):
            raise ValueError(
                "Using paged context fmha or fp8 context fmha requires requires paged kv cache"
            )
        if (
            self.build.plugin_configuration.use_fp8_context_fmha
            and not self.build.plugin_configuration.use_paged_context_fmha
        ):
            raise ValueError("Using fp8 context fmha requires paged context fmha")
        return self

    def _validate_hf_repo_id(self):
        try:
            validate_repo_id(self.build.checkpoint_repository.repo)
        except HFValidationError as e:
            raise ValueError(f"HuggingFace repository validation failed: {str(e)}") from e

    @property
    def requires_build(self):
        if self.build is not None:
            return True
        return False

    # TODO(Abu): Replace this with model_dump(json=True)
    # when pydantic v2 is used here
    def to_json_dict(self, verbose=True):
        return json.loads(self.json(exclude_unset=not verbose))


class TRTLLMSpeculativeDecodingConfiguration(BaseModel):
    target: TRTLLMConfiguration
    draft: TRTLLMConfiguration
    total_token_limit: int = 500000

    def __init__(self, **data):
        super().__init__(**data)
        self._spec_dec_configs = [
            self.target.build.speculative_decoding_mode,
            self.target.build.max_draft_len,
        ] + (
            [self.draft.runtime.num_draft_tokens]
            if self.draft.runtime and self.draft.runtime.num_draft_tokens
            else [False]
        )
        self._validate_spec_dec()

    def _validate_spec_dec(self):
        if any(self._spec_dec_configs):
            if not all(self._spec_dec_configs):
                raise ValueError(
                    "Speculative decoding requires all of `target.build.speculative_decoding_mode`, `target.build.max_draft_len`, and `draft.runtime.num_draft_tokens` to be configured."
                )
        for trt_llm_config in [self.target, self.draft]:
            if trt_llm_config.build.base_model is TrussTRTLLMModel.WHISPER:
                raise ValueError("Speculative decoding for Whisper is not supported.")
        if self.target.build.tensor_parallel_count != self.draft.build.tensor_parallel_count:
            raise ValueError(
                "Speculative decoding requires the same tensor parallelism for target and draft models."
            )

    def to_json_dict(self, verbose=True):
        return json.loads(self.json(exclude_unset=not verbose))



import logging
from typing import Optional

from briton.trtllm_config import TrussTRTLLMBatchSchedulerPolicy


def trtllm_config_check(config):
    if "trt_llm" not in config:
        raise ValueError("trt_llm config is required for this model")


def batch_scheduler_policy_to_int(
    policy: TrussTRTLLMBatchSchedulerPolicy, logger: logging.Logger
) -> int:
    if policy == TrussTRTLLMBatchSchedulerPolicy.MAX_UTILIZATION:
        return 0
    elif policy == TrussTRTLLMBatchSchedulerPolicy.GUARANTEED_NO_EVICT:
        return 1
    else:
        logger.warning(f"Unknown batch scheduler policy: {policy}. Using GUARANTEED_NO_EVICT.")
        return 1



from typing import Any, Awaitable, Callable, Dict, List, Optional

from briton.backend.backend_types import (
    InferBackend,
    LazyLoadParams,
    LoadParams,
    RequestDetails,
)
from briton.backend.briton_request import (
    model_input_to_briton_request,
    openai_spec_response,
)
from briton.data_structures import or_false
from briton.fsm_cache import FsmCache
from briton.schema import ModelInput


class DefaultBackend(InferBackend):
    """Regular OpenAI spec support backend."""

    def __init__(self):
        self._tokenizer = None
        self._briton_stub = None
        self._generate_request_id = None
        self._config = None

    def load(self, load_params: LoadParams) -> None:
        self._tokenizer = load_params.tokenizer
        self._generate_request_id = load_params.generate_request_id
        self._config = load_params.config

    async def lazy_load(self, lazy_load_params: LazyLoadParams) -> None:
        self._briton_stub = lazy_load_params.briton_stub

    async def accepts_request(self, model_input: ModelInput) -> Optional[RequestDetails]:
        if self._config is None or self._tokenizer is None:
            return None

        if not self._config.is_openai_compatible:
            return None

        input_ids = model_input.input_ids(self._tokenizer)
        return RequestDetails(input_ids=input_ids)

    async def infer(
        self,
        model_input: ModelInput,
        is_cancelled: Callable[[], Awaitable[bool]],
        add_schema_to_cache: Callable[[Dict[str, Any]], Awaitable[str]],
        request_details: RequestDetails,
    ) -> dict:
        # TODO(pankaj) Wire up request cancellation
        input_ids = request_details.input_ids
        request_id = self._generate_request_id()
        briton_request = await model_input_to_briton_request(
            request_id=request_id,
            model_input=model_input,
            input_ids=input_ids,
            tokenizer_eos_token_id=self._tokenizer.eos_token_id,
            tokenizer_pad_token_id=self._tokenizer.pad_token_id,
            add_schema_to_cache=add_schema_to_cache,
            default_max_tokens=self._config.default_max_tokens,
            max_seq_len=self._config.max_seq_len,
        )

        resp_iter = self._briton_stub.Infer(briton_request)

        eos_token = self._tokenizer.eos_token
        tool_call_token = self._config.tool_call_token
        return await self.generate_response(
            resp_iter=resp_iter,
            model_input=model_input,
            request_id=str(request_id),
            input_ids=input_ids,
            eos_token=eos_token,
            tool_call_token=tool_call_token,
        )

    async def generate_response(
        self,
        resp_iter,
        model_input: ModelInput,
        request_id: str,
        input_ids: List[int],
        eos_token: str,
        tool_call_token: str,
    ):
        streaming = or_false(model_input.stream)
        model_name = model_input.model if model_input.model else ""
        return await openai_spec_response(
            resp_iter=resp_iter,
            request_id=request_id,
            num_input_ids=len(input_ids),
            streaming=streaming,
            eos_token=eos_token,
            tool_call_token=tool_call_token,
            model_name=model_name,
            include_stream_usage=model_input.include_stream_usage,
            stop_words=model_input.stop,
            skip_special_tokens=model_input.skip_special_tokens,
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Awaitable, Callable, Dict, List, Optional

from pydantic import BaseModel
from transformers import PreTrainedTokenizerFast

from briton.proto import BritonStub
from briton.schema import ModelInput


class BatchSchedulerPolicy(IntEnum):
    MAX_UTILIZATION = 0
    GUARANTEED_NO_EVICT = 1


class BackendConfig(BaseModel):
    model_metadata: dict
    base_model: str
    tp: int = 1
    tokenizer_repo: str
    kv_cache_gpu_mem_fraction: float
    enable_kv_cache_reuse: bool
    enable_chunked_context: bool
    max_num_tokens: int
    max_seq_len: int

    batch_scheduler_policy: BatchSchedulerPolicy = BatchSchedulerPolicy.GUARANTEED_NO_EVICT
    default_max_tokens: Optional[int]
    max_fsm_workers: int
    tool_call_token: Optional[str]
    tool_call_token_id: Optional[int]

    is_openai_compatible: bool = True


@dataclass
class LoadParams:
    generate_request_id: Callable[[], str]
    tokenizer: PreTrainedTokenizerFast
    config: BackendConfig


@dataclass
class LazyLoadParams:
    briton_stub: BritonStub


class RequestDetails(BaseModel):
    input_ids: Optional[List[int]] = None
    num_input_ids: Optional[int] = None


class InferBackend(ABC):

    @abstractmethod
    def load(self, load_params: LoadParams):
        pass

    @abstractmethod
    async def lazy_load(self, lazy_load_params: LazyLoadParams):
        pass

    @abstractmethod
    async def accepts_request(self, model_input: ModelInput) -> Optional[RequestDetails]:
        """
        Whether this backend supports given request.

        :param model_input: model input
        :return: None if backend does not support the request,
            ParseRequestParams otherwise.
        """
        pass

    @abstractmethod
    async def infer(
        self,
        model_input: ModelInput,
        is_cancelled: Callable[[], Awaitable[bool]],
        add_schema_to_cache: Callable[[Dict[str, Any]], Awaitable[str]],
        request_details: RequestDetails,
    ) -> Any:
        """
        Similar to Truss's async predict function, this should return either a
        response(batch) or an async iterator(streaming). It should itself not be
        an async iterator.
        """
        pass



from .briton_pb2 import (
    AddedToken,
    AddedTokens,
    Batch,
    BritonConfig,
    InferenceRequest,
    StatesToTokens,
    TokenToNextState,
)
from .briton_pb2_grpc import BritonStub


__all__ = [
    "BritonStub",
    "StatesToTokens",
    "TokenToNextState",
    "InferenceRequest",
    "Batch",
    "BritonConfig",
    "AddedToken",
    "AddedTokens",
]



from typing import Any, Dict

from fastapi import HTTPException

from briton.constants import MODEL_INPUT_TO_BRITON_FIELD
from briton.proto import InferenceRequest


def set_briton_request_fields_from_model_input(
    model_input: Dict[str, Any], briton_request: InferenceRequest
):
    for model_input_key, briton_field in MODEL_INPUT_TO_BRITON_FIELD.items():
        if model_input_key in model_input:
            model_input_value = model_input[model_input_key]
            try:
                field_descriptor = briton_request.DESCRIPTOR.fields_by_name[briton_field]
                if field_descriptor.label == field_descriptor.LABEL_REPEATED:
                    getattr(briton_request, briton_field).extend(model_input_value)
                else:
                    try:
                        setattr(briton_request, briton_field, model_input_value)
                    except ValueError as e:
                        raise ValueError(
                            f"Error setting field {briton_field} with value {model_input_value}: {e}"
                        )
            except ValueError as e:
                raise HTTPException(status_code=400, detail=str(e))



import json
import logging
import random
import time
from json import JSONDecodeError
from typing import Any, AsyncGenerator, Callable, List, Literal, Optional

import openai.types.chat.chat_completion as chat_completion
import openai.types.chat.chat_completion_chunk as chat_completion_chunk
from fastapi import HTTPException
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)
from openai.types.completion_usage import CompletionUsage

from briton.constants import ALPHANUMERIC_CHARS

logger = logging.getLogger(__name__)


def _load_content_json(content: str) -> Any:
    """Safely load the content json from the input text."""
    try:
        return json.loads(content)
    except JSONDecodeError:
        raise HTTPException(status_code=400, detail="Tool call was cut off by max_tokens.")


def _generate_tool_call_id():
    """Mistral expects a length 9 alphanumeric id"""
    return "".join(random.choices(ALPHANUMERIC_CHARS, k=9))


def _create_tool_calls(content: str) -> List[ChatCompletionMessageToolCall]:
    content_json = _load_content_json(content)
    tool_calls = []
    for briton_fn in content_json:
        fn = Function(name=briton_fn["name"], arguments=json.dumps(briton_fn["parameters"]))
        tool_call = ChatCompletionMessageToolCall(
            id=_generate_tool_call_id(), function=fn, type="function"
        )
        tool_calls.append(tool_call)
    return tool_calls


def _finish_reason_from_text(
    text: str, eos_token: Optional[str] = None, stop_words: Optional[List[str]] = None
) -> Literal["stop", "length"]:
    if eos_token and text.endswith(eos_token):
        return "stop"
    if stop_words and text.endswith(tuple(stop_words)):
        return "stop"
    return "length"


def remove_suffix_from_text(
    text: str,
    eos_token: Optional[str] = None,
    stop_words: Optional[List[str]] = None,
    skip_special_tokens: Optional[List[str]] = None,
) -> str:
    if eos_token and text.endswith(eos_token):
        return text.removesuffix(eos_token)
    if stop_words:
        for stop_word in stop_words:
            if text.endswith(stop_word):
                return text.removesuffix(stop_word)
    # HACK (bdubayah): this could end up being very expensive.
    if skip_special_tokens:
        for special_token in skip_special_tokens:
            text = text.replace(special_token, "")
    return text


def create_completion(
    req_id: str,
    model: str,
    input_text: str,
    eos_token: Optional[str] = None,
    tool_token: Optional[str] = None,
    prompt_tokens: Optional[int] = None,
    completion_tokens: Optional[int] = None,
    stop_words: Optional[List[str]] = None,
    skip_special_tokens: Optional[List[str]] = None,
) -> ChatCompletion:
    created = int(time.time())
    finish_reason = _finish_reason_from_text(input_text, eos_token, stop_words)
    content = remove_suffix_from_text(input_text, eos_token, stop_words, skip_special_tokens)
    tool_calls = None
    if tool_token is not None and content.startswith(tool_token):
        finish_reason = "tool_calls"
        content = content.removeprefix(tool_token)
        tool_calls = _create_tool_calls(content)
        content = None
    message = ChatCompletionMessage(content=content, role="assistant", tool_calls=tool_calls)
    choice = chat_completion.Choice(finish_reason=finish_reason, index=0, message=message)
    usage = None
    if prompt_tokens is not None and completion_tokens is not None:
        usage = CompletionUsage(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=prompt_tokens + completion_tokens,
        )
    return ChatCompletion(
        id=req_id,
        choices=[choice],
        created=created,
        model=model,
        object="chat.completion",
        usage=usage,
    )


def _make_sse_chunk(chunk: ChatCompletionChunk) -> str:
    return f"data: {chunk.model_dump_json()}\n\n"


# It's the responsibility of the caller to pass the full content
def _create_tool_call_deltas(
    content: str,
) -> List[chat_completion_chunk.ChoiceDeltaToolCall]:
    content_json = _load_content_json(content)
    tool_call_deltas = []
    for i, briton_fn in enumerate(content_json):
        if not (isinstance(briton_fn, dict) and "name" in briton_fn and "parameters" in briton_fn):
            logger.error(f"Generated tool calls {content_json} are not valid")
            continue
        fn = chat_completion_chunk.ChoiceDeltaToolCallFunction(
            name=briton_fn["name"], arguments=json.dumps(briton_fn["parameters"])
        )
        tool_call_delta = chat_completion_chunk.ChoiceDeltaToolCall(
            index=i, id=_generate_tool_call_id(), function=fn, type="function"
        )
        tool_call_deltas.append(tool_call_delta)
    return tool_call_deltas


def _create_completion_chunk(
    id: str,
    created: int,
    model: str,
    content: Optional[str] = None,
    role: Optional[Literal["system", "user", "assistant", "tool"]] = None,
    finish_reason: Optional[
        Literal["stop", "length", "tool_calls", "content_filter", "function_call"]
    ] = None,
    tool_calls: Optional[List[chat_completion_chunk.ChoiceDeltaToolCall]] = None,
    usage: Optional[CompletionUsage] = None,
) -> ChatCompletionChunk:
    delta = chat_completion_chunk.ChoiceDelta(content=content, role=role, tool_calls=tool_calls)
    choice = chat_completion_chunk.Choice(index=0, delta=delta, finish_reason=finish_reason)
    return ChatCompletionChunk(
        id=id,
        choices=[choice],
        created=created,
        model=model,
        object="chat.completion.chunk",
        usage=usage,
    )


async def create_completion_chunks(
    req_id: str,
    model: str,
    input_text: AsyncGenerator[str, None],
    eos_token: Optional[str] = None,
    tool_token: Optional[str] = None,
    prompt_tokens: Optional[int] = None,
    completion_tokens_fn: Optional[Callable[[], int]] = None,
    stop_words: Optional[List[str]] = None,
    skip_special_tokens: Optional[List[str]] = None,
) -> AsyncGenerator[str, None]:
    created = int(time.time())
    start_chunk = _create_completion_chunk(
        id=req_id, created=created, model=model, content="", role="assistant"
    )
    is_first_iter = True
    delta = None
    async for delta in input_text:
        if is_first_iter:
            if tool_token is not None and delta.startswith(tool_token):
                break
            is_first_iter = False
            yield _make_sse_chunk(start_chunk)
        content = remove_suffix_from_text(delta, eos_token, stop_words, skip_special_tokens)
        if len(content) == 0:
            continue  # Don't send empty chunks
        chunk = _create_completion_chunk(id=req_id, created=created, model=model, content=content)
        yield _make_sse_chunk(chunk)

    # Handle function call case
    if is_first_iter and delta is not None and tool_token and delta.startswith(tool_token):
        full_text = delta.removeprefix(tool_token)
        async for delta in input_text:
            full_text += delta
        tool_calls = _create_tool_call_deltas(
            remove_suffix_from_text(full_text, eos_token, stop_words, skip_special_tokens)
        )
        chunk = _create_completion_chunk(
            id=req_id, created=created, model=model, tool_calls=tool_calls
        )
        yield _make_sse_chunk(start_chunk)
        yield _make_sse_chunk(chunk)
        finish_reason = "tool_calls"
    else:
        finish_reason = (
            _finish_reason_from_text(delta, eos_token, stop_words)
            if delta is not None
            else "length"
        )
    end_chunk = _create_completion_chunk(
        id=req_id, created=created, model=model, finish_reason=finish_reason
    )
    yield _make_sse_chunk(end_chunk)
    if prompt_tokens is not None and completion_tokens_fn is not None:
        completion_tokens = completion_tokens_fn()
        usage = CompletionUsage(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=prompt_tokens + completion_tokens,
        )
        usage_chunk = _create_completion_chunk(id=req_id, created=created, model=model, usage=usage)
        yield _make_sse_chunk(usage_chunk)
    yield "data: [DONE]\n\n"



import os
from itertools import count


class RequestIdGenerator:
    def __init__(self):
        self._counter = count(start=1)

    def __call__(self) -> int:
        """Calculate unique request id.

        Not thread safe, but safe to use in single threaded async context. There
        are no async operations here, so this function is unlikely to be
        preempted in the middle. This is important otherwise we may end up with
        duplicate ids.
        """
        next_id = next(self._counter)
        return int(str(os.getpid()) + str(next_id))
